/*
 * FSSetTarget.java
 * Transform
 * 
 * Copyright (c) 2001-2006 Flagstone Software Ltd. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation 
 *    and/or other materials provided with the distribution.
 *  * Neither the name of Flagstone Software Ltd. nor the names of its contributors 
 *    may be used to endorse or promote products derived from this software 
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package com.flagstone.transform;

/**
FSSetTarget is used to control the time-line of a particular movie clip. 
 
<p>The action performs a "context switch". All following actions such as GotoFrame, Play, etc. will be applied to the specified object until another <b>FSSetTarget</b> action is executed. Setting the target to be the empty string ("") returns the target to the movie's main timeline.</p>

<table class="datasheet">

<tr><th align="left" colspan="2">Attributes</th></tr>

<tr>
<td><a name="FSSetTarget_0">type</a></td>
<td>Identifies the action when it is encoded. Read-only.</td>
</tr>

<tr>
<td><a name="FSSetTarget_1">target</a></td>
<td>The name assigned to a movie clip. Names are assigned using the FSPlaceObject2 class when the movie clip is placed on the Flash Player's Display List.</td>
</tr>

</table>

<h1 class="datasheet">Example</h1>

<pre>
// Direct all subsequent actions to this movie clip

actions.add(new FSSetTarget("/MovieClip:3"));
...
&lt;movie clip actions&gt;
...

// Restore the actions to the main movie timeline

actions.add(new FSSetTarget(""));
</pre>


<h1 class="datasheet">History</h1>

<p>The FSSetTarget class represents the SetTarget action of the Macromedia Flash (SWF) File Format Specification. It was introduced in Flash 3.</p>
 */ 
public class FSSetTarget extends FSActionObject
{
    private String target = "";
    
    /**
     * Construct an FSSetTarget object, initalizing it with values decoded from
     * an encoded object.
     * 
     * @param coder an FSCoder containing the binary data.
     */
    public FSSetTarget(FSCoder coder)
    {
        super(SetTarget);
        decode(coder);
    }
    /** Constructs an FSSetTarget action that changes the context to the specified target.

        @param aString the name of a movie clip.
        */
    public FSSetTarget(String aString)
    {
        super(SetTarget);
        setTarget(aString);
    }
    /**
     * Constructs an FSSetTarget object by copying values from an existing 
     * object.
     *
     * @param obj an FSSetTarget object.
     */
    public FSSetTarget(FSSetTarget obj)
    {
        super(obj);
        target = new String(obj.target);
    }    

    /** Gets the name of the target movie clip. 

        @return the name of the target movie clip.
        */
    public String getTarget() { return target; }

    /** Sets the name of the target movie clip.

        @param aString the name of a movie clip.
        */
    public void setTarget(String aString)
    {
        target = aString;
    } 

    public boolean equals(Object anObject)
    {
        boolean result = false;
        
        if (super.equals(anObject))
            result = target.equals(((FSSetTarget)anObject).getTarget());

        return result;
    }

    public void appendDescription(StringBuffer buffer, int depth)
    {
        buffer.append(name());
        
        if (depth > 0)
        {
            buffer.append(": { ");
            Transform.append(buffer, "target", target);
            buffer.append("}");
        }
    }

    public int length(FSCoder coder)
    {
        super.length(coder);
        
        length += coder.strlen(target, true);
        
        return length;
    }
    
    public void encode(FSCoder coder)
    {
        super.encode(coder);

         coder.writeString(target);
        coder.writeWord(0, 1);
        coder.endObject(name());
    }
    
    public void decode(FSCoder coder)
    {
        super.decode(coder);
        target = coder.readString();
        coder.endObject(name());
    }
}