/*
 * FSTabOrder.java
 * Transform
 * 
 * Copyright (c) 2001-2006 Flagstone Software Ltd. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation 
 *    and/or other materials provided with the distribution.
 *  * Neither the name of Flagstone Software Ltd. nor the names of its contributors 
 *    may be used to endorse or promote products derived from this software 
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package com.flagstone.transform;

/**
The FSTabOrder class is used to set the tabbing order of text fields, movie clips and 
buttons visible on the display list.
 
<table class="datasheet">

<tr><th align="left" colspan="2">Attributes</th></tr>

<tr>
<td><a name="FSDefineVideo_0">type</a></td>
<td>Identifies the data structure when it is encoded. Read-only.</td>
</tr>

<tr>
<td><a name="FSDefineVideo_1">layer</a></td>
<td>The layer number which contains the object assigned to the tabbing order.</td>
</tr>

<tr>
<td><a name="FSDefineVideo_2">index</a></td>
<td>The index of the object in the tabbing order.</td>
</tr>

</table>

<h1 class="datasheet">History</h1>

<p>The FSTabOrder object represents the SetTabIndex tag in the Flash file specification . It was introduced in Flash 7.</p>
 */  
public class FSTabOrder extends FSMovieObject
{
    private int layer = 0; 
    private int index = 0; 
    
    /**
     * Construct an FSTabOrder object, initalizing it with values decoded from
     * an encoded object.
     * 
     * @param coder an FSCoder containing the binary data.
     */
    public FSTabOrder(FSCoder coder)
    {
        super(TabOrder);
        decode(coder);
    }

    /**
     * Construct a FSTabOrder object that set the tab order for the object on the display list
     * at the specified layer.
     *
     * @param layer the layer number which contains the object assigned to the tabbing order.
     * @param index the index of the object in the tabbing order.
     */
    public FSTabOrder(int layer, int index)
    {
        super(TabOrder);
        setLayer(layer);
        setIndex(index);
    }
    /**
     * Constructs an FSTabOrder object by copying values from an existing object.
     *
     * @param obj an FSTabOrder object.
     */
    public FSTabOrder(FSTabOrder obj)
    {
        super(obj);
        
        layer = obj.layer;
        index = obj.index;
    }
    /**
     * Gets the layer number which contains the object assigned to the tabbing order.
     *
     * @return the layer number.
     */
    public int getLayer() 
    {
        return layer;
    }

    /**
     * Sets the layer number which contains the object assigned to the tabbing order.
     *
     * @param layer the layer number.
     */
    public void setLayer(int layer)
    {
        this.layer = layer;
    }
    
    /**
     * Gets the index of the object in the tabbing order.
     *
     * @return the index in the tabbing order.
     */
    public int getIndex() 
    {
        return index;
    }

    /**
     * Sets the index of the object in the tabbing order.
     *
     * @param index the index in the tabbing order.
     */
    public void setIndex(int index)
    {
        this.index = index;
    }
    
    public boolean equals(Object anObject)
    {
        boolean result = false;
        
        if (super.equals(anObject))
        {
            FSTabOrder typedObject = (FSTabOrder) anObject;
            
            result = layer == typedObject.layer;
            result = result && index == typedObject.index;
        }
        return result;
    }

    public void appendDescription(StringBuffer buffer, int depth)
    {
        buffer.append(name());
        
        if (depth > 0)
        {
            buffer.append(": { ");
            Transform.append(buffer, "layer", this.layer);
            Transform.append(buffer, "index", index);
              buffer.append("}");
        }
    }

    public int length(FSCoder coder)
    {
        super.length(coder);
        
        length += 4;
        
        return length;
    }
    
    public void encode(FSCoder coder)
    {
        super.encode(coder);
        
        coder.writeWord(layer, 2);
        coder.writeWord(index, 2);
        coder.endObject(name());
    }
    
    public void decode(FSCoder coder)
    {
        super.decode(coder);
        
        layer = coder.readWord(2, false);
        index = coder.readWord(2, false);
        coder.endObject(name());
    }
}
